# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shlex
from typing import Any, Dict
from typing import Sequence  # noqa: F401

from snapcraft.internal.errors import SnapcraftError as _SnapcraftError


class ProviderBaseError(_SnapcraftError):
    """Base Exception for all provider related exceptions."""


class ProviderNotSupportedError(ProviderBaseError):

    fmt = (
        "The {provider!r} provider is not supported, please choose a "
        "different one and try again."
    )

    def __init__(self, *, provider: str) -> None:
        super().__init__(provider=provider)


class ProviderCommandNotFound(ProviderBaseError):

    fmt = (
        "{command!r} command not found: this command is necessary to build in "
        "this environment.\n"
        "Install {command!r} or if already installed, ensure it is "
        "on the system PATH, and try again."
    )

    def __init__(self, *, command: str) -> None:
        super().__init__(command=command)


class _GenericProviderError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to {action} the instance with "
        "{provider_name!r}: returned exit code {exit_code!r}.\n"
        "Ensure that {provider_name!r} is setup correctly and try again."
    )


class ProviderCommunicationError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to communicate with the instance "
        "using {protocol!r} over port {port}: {error}."
    )

    def __init__(self, *, protocol: str, port: int, error: str) -> None:
        super().__init__(protocol=protocol, port=port, error=error)


class ProviderLaunchError(_GenericProviderError):
    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(
            action="launch", provider_name=provider_name, exit_code=exit_code
        )


class ProviderStartError(_GenericProviderError):
    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(
            action="start", provider_name=provider_name, exit_code=exit_code
        )


class ProviderStopError(_GenericProviderError):
    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(
            action="stop", provider_name=provider_name, exit_code=exit_code
        )


class ProviderDeleteError(_GenericProviderError):
    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(
            action="delete", provider_name=provider_name, exit_code=exit_code
        )


class ProviderExecError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to execute {command_string!r} with "
        "{provider_name!r}: returned exit code {exit_code!r}."
    )

    def __init__(
        self, *, provider_name: str, command: Sequence[str], exit_code: int
    ) -> None:
        command_string = " ".join(shlex.quote(i) for i in command)
        super().__init__(
            provider_name=provider_name,
            command=command,
            command_string=command_string,
            exit_code=exit_code,
        )


class ProviderShellError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to provide a shell with "
        "{provider_name!r}: returned exit code {exit_code!r}."
    )

    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(provider_name=provider_name, exit_code=exit_code)


class ProviderMountError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to mount using {provider_name!r}: "
        "returned exit code {exit_code!r}."
    )

    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(provider_name=provider_name, exit_code=exit_code)


class ProviderUnMountError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to unmount using {provider_name!r}: "
        "returned exit code {exit_code!r}."
    )

    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(provider_name=provider_name, exit_code=exit_code)


class ProviderFileCopyError(ProviderBaseError):

    fmt = (
        "An error occurred when trying to copy files using {provider_name!r}: "
        "returned exit code {exit_code!r}."
    )

    def __init__(self, *, provider_name: str, exit_code: int) -> None:
        super().__init__(provider_name=provider_name, exit_code=exit_code)


class ProviderInfoError(ProviderBaseError):

    fmt = (
        "An error occurred when using {provider_name!r} to "
        "query the status of the instance: returned exit code {exit_code!r}: {stderr!s}."
    )

    def __init__(self, *, provider_name: str, exit_code: int, stderr: bytes) -> None:
        super().__init__(
            provider_name=provider_name, exit_code=exit_code, stderr=stderr.decode()
        )


class ProviderInstanceNotFoundError(ProviderBaseError):

    fmt = "Cannot find an instance named {instance_name!r}."

    def __init__(self, *, instance_name: str) -> None:
        super().__init__(instance_name=instance_name)


class ProviderInfoDataKeyError(ProviderBaseError):

    fmt = (
        "The data returned by {provider_name!r} was not expected. "
        "It is missing a required key {missing_key!r} in {data!r}."
    )

    def __init__(
        self, *, provider_name: str, missing_key: str, data: Dict[str, Any]
    ) -> None:
        super().__init__(
            provider_name=provider_name, missing_key=missing_key, data=data
        )


class ProviderBadDataError(ProviderBaseError):

    fmt = (
        "The data returned by {provider_name!r} was not expected "
        "or in the wrong format: {data!r}."
    )

    def __init__(self, *, provider_name: str, data: str) -> None:
        super().__init__(provider_name=provider_name, data=data)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.util
import os
import sys
import tempfile
import traceback
from textwrap import dedent
from typing import Dict  # noqa: F401

import click

from . import echo
import snapcraft
from snapcraft.config import CLIConfig as _CLIConfig
from snapcraft.internal import errors

# raven is not available on 16.04
try:
    from raven import Client as RavenClient
    from raven.transport import RequestsHTTPTransport
except ImportError:
    RavenClient = None

# TODO:
# - annotate the part and lifecycle step in the message
# - add link to privacy policy
_MSG_TRACEBACK_PRINT = "Sorry, an error occurred in Snapcraft:"
_MSG_TRACEBACK_FILE = "Sorry, an error occurred in Snapcraft."
_MSG_SEND_TO_SENTRY_TRACEBACK_PROMPT = dedent(
    """\
    We would appreciate it if you anonymously reported this issue.
    No other data than the traceback ({!r}) and the version of snapcraft in use will be sent.
    Would you like to send this error data? (Yes/No/Always)"""
)
_MSG_SEND_TO_SENTRY_THANKS = "Thank you, sent."
_MSG_SILENT_REPORT = (
    "Sending an error report because SNAPCRAFT_ENABLE_SILENT_REPORT is set."
)
_MSG_ALWAYS_REPORT = dedent(
    """\
    Sending an error report because ALWAYS was selected in a past prompt.
    This behavior can be changed by changing the always_send entry in {}."""
)
_MSG_MANUALLY_REPORT = dedent(
    """\
    We would appreciate it if you created a bug report at
    https://launchpad.net/snapcraft/+filebug with the above text included."""
)

_YES_VALUES = ["yes", "y"]
_NO_VALUES = ["no", "n"]
_ALWAYS_VALUES = ["always", "a"]


def exception_handler(  # noqa: C901
    exception_type, exception, exception_traceback, *, debug=False
) -> None:
    """Catch all Snapcraft exceptions unless debugging.

    This function is the global excepthook, properly handling uncaught
    exceptions and determine if they need to be reported.

    These are the rules of engagement:

        - a non snapcraft handled error occurs and raven is setup,
          so we go over confirmation logic showing the relevant traceback
        - a non snapcraft handled error occurs and raven is not setup,
          so we just show the traceback
        - a snapcraft handled error occurs, debug=True so a traceback
          is shown
        - a snapcraft handled error occurs, debug=False so only the
          exception message is shown
    """
    # TODO pickle the traceback if on a manged host so the actual host can deal with it.
    exc_info = (exception_type, exception, exception_traceback)
    exit_code = 1
    is_snapcraft_error = issubclass(exception_type, errors.SnapcraftError)
    is_snapcraft_reportable_error = issubclass(
        exception_type, errors.SnapcraftReportableError
    )
    is_raven_setup = RavenClient is not None
    is_snapcraft_managed_host = (
        distutils.util.strtobool(os.getenv("SNAPCRAFT_MANAGED_HOST", "n")) == 1
    )
    is_connected_to_tty = sys.stdout.isatty()
    ask_to_report = False

    if not is_snapcraft_error:
        ask_to_report = True
    elif is_snapcraft_error and debug:
        exit_code = exception.get_exit_code()
        traceback.print_exception(*exc_info)
    elif is_snapcraft_error and not debug:
        exit_code = exception.get_exit_code()
        echo.error(str(exception))
        if is_snapcraft_reportable_error:
            ask_to_report = True
    else:
        click.echo("Unhandled error case")
        exit_code = -1

    if ask_to_report:
        if not is_raven_setup or is_snapcraft_managed_host:
            click.echo(_MSG_TRACEBACK_PRINT)
            traceback.print_exception(*exc_info, file=sys.stdout)
            click.echo(_MSG_MANUALLY_REPORT)
        else:
            if is_connected_to_tty:
                click.echo(_MSG_TRACEBACK_FILE)
            else:
                click.echo(_MSG_TRACEBACK_PRINT)
                traceback.print_exception(*exc_info, file=sys.stdout)
            trace_filepath = _handle_trace_output(exc_info)
            if _is_send_to_sentry(trace_filepath):
                try:
                    _submit_trace(exc_info)
                except Exception as exc:
                    # we really cannot do much more from this exit handler.
                    echo.error(
                        "Encountered an issue while trying to submit the report: {}".format(
                            str(exc)
                        )
                    )
                else:
                    click.echo(_MSG_SEND_TO_SENTRY_THANKS)

    sys.exit(exit_code)


def _handle_trace_output(exc_info) -> str:
    trace_filepath = os.path.join(tempfile.mkdtemp(), "trace.txt")
    with open(trace_filepath, "w") as trace_file:
        # mypy does not like *exc_info with a kwarg that follows
        # snapcraft/cli/_errors.py:132: error: "print_exception" gets multiple values for keyword argument "file"
        traceback.print_exception(
            exc_info[0], exc_info[1], exc_info[2], file=trace_file
        )
    return trace_filepath


def _is_send_to_sentry(trace_filepath: str) -> bool:  # noqa: C901
    # Check to see if error reporting has been disabled
    if (
        distutils.util.strtobool(os.getenv("SNAPCRAFT_ENABLE_ERROR_REPORTING", "y"))
        == 0
    ):
        return False

    # Check the environment to see if we should allow for silent reporting
    if distutils.util.strtobool(os.getenv("SNAPCRAFT_ENABLE_SILENT_REPORT", "n")) == 1:
        click.echo(_MSG_SILENT_REPORT)
        return True

    # If ALWAYS has already been selected from before do not even bother to
    # prompt again.
    config_errors = None
    try:
        with _CLIConfig(read_only=True) as cli_config:
            if cli_config.get_sentry_send_always():
                click.echo(_MSG_ALWAYS_REPORT.format(cli_config.config_path))
                return True
    except errors.SnapcraftInvalidCLIConfigError as config_error:
        echo.warning(
            "There was an issue while trying to retrieve "
            "configuration data: {!s}".format(config_error)
        )
        config_errors = config_error

    # Either ALWAYS has not been selected in a previous run or the
    # configuration for where that value is stored cannot be read, so
    # resort to prompting.
    try:
        response = _prompt_sentry(trace_filepath)
    except click.exceptions.Abort:
        # This was most likely triggered by a KeyboardInterrupt so
        # adding a new line makes things look nice.
        print()
        return False
    if response in _YES_VALUES:
        return True
    elif response in _NO_VALUES:
        return False
    elif response in _ALWAYS_VALUES:
        if config_errors is not None:
            echo.warning(
                "Not saving choice to always send data to Sentry as "
                "the configuration file is corrupted.\n"
                "Please edit and fix or alternatively remove the "
                "configuration file {!r} from disk.".format(config_errors.config_file)
            )  # type: ignore
        else:
            click.echo("Saving choice to always send data to Sentry")
            with _CLIConfig() as cli_config:
                cli_config.set_sentry_send_always(True)
        return True
    else:
        echo.warning("Could not determine choice, assuming no.")
        return False


def _prompt_sentry(trace_filepath: str):
    msg = _MSG_SEND_TO_SENTRY_TRACEBACK_PROMPT.format(trace_filepath)
    all_valid = _YES_VALUES + _NO_VALUES + _ALWAYS_VALUES

    def validate(value):
        if value.lower() in all_valid:
            return value
        raise click.BadParameter("Please choose a valid answer.")

    return click.prompt(msg, default="no", value_proc=validate).lower()


def _submit_trace(exc_info):
    kwargs = dict()  # Dict[str,str]
    if "+git" not in snapcraft.__version__:
        kwargs["release"] = snapcraft.__version__

    client = RavenClient(
        "https://b0fef3e0ced2443c92143ae0d038b0a4:"
        "b7c67d7fa4ee46caae12b29a80594c54@sentry.io/277754",
        transport=RequestsHTTPTransport,
        # Should Raven automatically log frame stacks (including locals)
        # for all calls as it would for exceptions.
        auto_log_stacks=False,
        # Set a name to not send the real hostname.
        name="snapcraft",
        # Removes all stacktrace context variables. This will cripple the
        # functionality of Sentry, as you’ll only get raw tracebacks,
        # but it will ensure no local scoped information is available to the
        # server.
        processors=(
            "raven.processors.RemoveStackLocalsProcessor",
            "raven.processors.SanitizePasswordsProcessor",
        ),
        **kwargs
    )
    client.captureException(exc_info=exc_info)



#!/usr/bin/python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from snapcraft.internal.indicators import is_dumb_terminal


class _StdoutFilter(logging.Filter):
    def filter(self, record):
        return record.levelno <= logging.WARNING


class _StderrFilter(logging.Filter):
    def filter(self, record):
        return record.levelno >= logging.ERROR


class _ColoredFormatter(logging.Formatter):
    RESET = "\033[0m"
    LEVEL_COLORS = {
        "INFO": "\033[0;32m",  # Green
        "WARNING": "\033[1;33m",  # Yellow
        "ERROR": "\033[0;31m",  # Dark red
        "CRITICAL": "\033[1;31m",  # Light red
    }

    def format(self, record):
        color = self.LEVEL_COLORS.get(record.levelname, None)
        log_message = super().format(record)
        if color:
            return "{color}{message}{reset}".format(
                color=color, message=log_message, reset=self.RESET
            )

        return log_message


def configure(logger_name=None, log_level=None):
    if not log_level:
        log_level = logging.INFO

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.addFilter(_StdoutFilter())
    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.addFilter(_StderrFilter())
    handlers = [stdout_handler, stderr_handler]

    if is_dumb_terminal():
        formatter = logging.Formatter(style="{")
    else:
        formatter = _ColoredFormatter(style="{")
    logger = logging.getLogger(logger_name)
    for handler in handlers:
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(log_level)

    # INFO by default for the requests lib as it is too noisy
    if log_level == logging.DEBUG:
        logging.getLogger("requests").setLevel(log_level)
    else:
        logging.getLogger("requests").setLevel(logging.WARNING)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Snapcraft integration for Travis (CI).

This is an *EXPERIMENTAL* feature and subject to incompatible changes in
the future, please use with caution.

This command currently depends on a working `travis` CLI environment and
a previously initialized Travis project (`.travis.yml`).

Make sure your Travis project is also configured to "Build pushes", this
way every new push to `master` will result in a new snap revision in the
Store.

This operation will acquire properly attenuated Store credentials and
encrypt them for use in your testbed (`.snapcraft/travis_snapcraft.cfg`),
only Travis has the private key to decrypt it and will be only available
to branches of the same repository, not forks.

Then it will adjust Travis configuration ('.travis.yml') with the commands
to decrypt credentials during 'after_success' phase and install latest
`snapcraft` to build and release your snap (inside a snapcore/snapcraft docker
container) during the 'deploy' phase.

See the example below::

    sudo: required
    services:
    - docker
    after_success:
    - openssl aes-256-cbc -K <travis-key> -iv <travis-iv>
      -in .snapcraft/travis_snapcraft.cfg
      -out .snapcraft/snapcraft.cfg -d
    deploy:
      skip_cleanup: true
      provider: script
      script: docker run -v $(pwd):$(pwd) -t snapcore/snapcraft sh -c
        "apt update -qq && cd $(pwd) &&
        snapcraft && snapcraft push *.snap --release edge"
      on:
        branch: master

The dedicated credentials will be functional for exactly 1 year or until
a password change on the related Ubuntu One SSO account. In order to
refresh the project credentials, please run the following command::

    $ snapcraft enable-ci travis --refresh
"""
from contextlib import ExitStack, contextmanager
import logging
import os
import subprocess
import tempfile

from snapcraft import storeapi, yaml_utils
from snapcraft.file_utils import requires_command_success, requires_path_exists
from snapcraft.internal import project_loader
from snapcraft._store import login
from snapcraft.config import LOCAL_CONFIG_FILENAME

logger = logging.getLogger(__name__)

TRAVIS_CONFIG_FILENAME = ".travis.yml"
ENCRYPTED_CONFIG_FILENAME = ".snapcraft/travis_snapcraft.cfg"


class TravisRuntimeError(Exception):
    """Local runtime error to not confuse importlib.

    Raising `RuntimeError` from a module imported in runtime
    results in python hanging instead of just exiting.
    """


def _acquire_and_encrypt_credentials(packages, channels):
    """Acquire and encrypt Store credentials for Travis jobs."""
    # XXX cprov 20161116: Needs caveat syntax for restricting origins
    # (IP or reverse-dns) but Travis sudo-enabled containers, needed for
    # running xenial snapcraft, do not have static egress routes.
    # See https://docs.travis-ci.com/user/ip-addresses.
    logger.info("Acquiring specific authorization information ...")
    store = storeapi.StoreClient()
    # Travis cannot register new names or add new developers.
    acls = ["package_access", "package_push", "package_release"]
    if not login(
        store=store, acls=acls, packages=packages, channels=channels, save=False
    ):
        raise TravisRuntimeError("Cannot continue without logging in successfully.")

    logger.info(
        "Encrypting authorization for Travis and adjusting project to "
        'automatically decrypt and use it during "after_success".'
    )
    with tempfile.NamedTemporaryFile(mode="w") as fd:
        store.conf.save(config_fd=fd)
        fd.flush()
        os.makedirs(os.path.dirname(LOCAL_CONFIG_FILENAME), exist_ok=True)
        cmd = [
            "travis",
            "encrypt-file",
            "--force",
            "--add",
            "after_success",
            "--decrypt-to",
            LOCAL_CONFIG_FILENAME,
            fd.name,
            ENCRYPTED_CONFIG_FILENAME,
        ]
        try:
            subprocess.check_output(cmd, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as err:
            raise TravisRuntimeError(
                "`travis encrypt-file` failed: {}\n{}".format(
                    err.returncode, err.stderr.decode()
                )
            )


@contextmanager
def requires_travis_preconditions():
    """Verify all Travis CI integration preconditions."""
    required = (
        requires_command_success(
            "travis settings",
            not_found_fmt=(
                "Travis CLI (`{cmd_list[0]}`) is not available.\n"
                "Please install it before trying this command again:\n\n"
                "    $ sudo apt install ruby-dev ruby-ffi libffi-dev\n"
                "    $ sudo gem install travis\n"
            ),
            failure_fmt=(
                "Travis CLI (`{command}`) is not functional or you are not "
                "allowed to access this repository settings.\n"
                "Make sure it works correctly in your system before trying "
                "this command again."
            ),
        ),
        requires_command_success(
            "git status",
            not_found_fmt=(
                "Git (`{cmd_list[0]}`) is not available, this tool cannot "
                "verify its prerequisites.\n"
                "Please install it before trying this command again:\n\n"
                "    $ sudo apt install git\n"
            ),
            failure_fmt=(
                "The current directory is not a Git repository.\n"
                "Please switch to the desired project repository where "
                "Travis should be enabled."
            ),
        ),
        requires_path_exists(
            TRAVIS_CONFIG_FILENAME,
            error_fmt=(
                "Travis project is not initialized for the current "
                "directory.\n"
                "Please initialize Travis project (e.g. `travis init`) with "
                "appropriate parameters."
            ),
        ),
    )
    with ExitStack() as cm:
        [cm.enter_context(c) for c in required]
        yield


@requires_travis_preconditions()
def refresh(project):
    series = storeapi.constants.DEFAULT_SERIES
    project_config = project_loader.load_config(project)
    snap_name = project_config.data["name"]
    logger.info(
        'Refreshing credentials to push and release "{}" snaps '
        "to edge channel in series {}".format(snap_name, series)
    )

    packages = [{"name": snap_name, "series": series}]
    channels = ["edge"]
    _acquire_and_encrypt_credentials(packages, channels)

    logger.info(
        "Done. Please commit the changes to `{}` file.".format(
            ENCRYPTED_CONFIG_FILENAME
        )
    )


@requires_travis_preconditions()
def enable(project):
    series = storeapi.constants.DEFAULT_SERIES
    project_config = project_loader.load_config(project)
    snap_name = project_config.data["name"]
    logger.info(
        "Enabling Travis testbeds to push and release {!r} snaps "
        "to edge channel in series {!r}".format(snap_name, series)
    )

    packages = [{"name": snap_name, "series": series}]
    channels = ["edge"]
    _acquire_and_encrypt_credentials(packages, channels)

    logger.info(
        'Configuring "deploy" phase to build and release the snap in the ' "Store."
    )
    with open(TRAVIS_CONFIG_FILENAME, "r+") as fd:
        travis_conf = yaml_utils.load(fd)
        # Enable 'sudo' capability and 'docker' service.
        travis_conf["sudo"] = "required"
        services = travis_conf.setdefault("services", [])
        if "docker" not in services:
            services.append("docker")
        # Add a 'deploy' section with 'script' provider for building and
        # release the snap within a xenial docker container.
        travis_conf["deploy"] = {
            "skip_cleanup": True,
            "provider": "script",
            "script": (
                "docker run -v $(pwd):$(pwd) -t snapcore/snapcraft sh -c "
                '"apt update -qq && cd $(pwd) && '
                'snapcraft && snapcraft push *.snap --release edge"'
            ),
            "on": {"branch": "master"},
        }
        fd.seek(0)
        yaml_utils.dump(travis_conf, stream=fd)

    logger.info(
        "Done. Now you just have to review and commit changes in your "
        "Travis project (`{}`).\n"
        "Also make sure you add the new `{}` file.".format(
            TRAVIS_CONFIG_FILENAME, ENCRYPTED_CONFIG_FILENAME
        )
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Data/methods shared between plugins and snapcraft
import glob
import logging
import math
import os
import shlex
import shutil
import subprocess
import sys
import tempfile
import urllib
from contextlib import suppress
from typing import Callable, List

from snapcraft.internal import errors


SNAPCRAFT_FILES = [
    "snapcraft.yaml",
    ".snapcraft.yaml",
    ".snapcraft",
    "parts",
    "stage",
    "prime",
    "snap",
]
_DEFAULT_PLUGINDIR = os.path.join(sys.prefix, "share", "snapcraft", "plugins")
_plugindir = _DEFAULT_PLUGINDIR
_DEFAULT_SCHEMADIR = os.path.join(sys.prefix, "share", "snapcraft", "schema")
_schemadir = _DEFAULT_SCHEMADIR
_DEFAULT_EXTENSIONSDIR = os.path.join(sys.prefix, "share", "snapcraft", "extensions")
_extensionsdir = _DEFAULT_EXTENSIONSDIR
_DEFAULT_LEGACY_SNAPCRAFT_DIR = os.path.join(sys.prefix, "legacy_snapcraft")
_legacy_snapcraft_dir = _DEFAULT_LEGACY_SNAPCRAFT_DIR
_DOCKERENV_FILE = "/.dockerenv"

MAX_CHARACTERS_WRAP = 120

env = []  # type: List[str]

logger = logging.getLogger(__name__)


def assemble_env():
    return "\n".join(["export " + e for e in env])


def _run(cmd: List[str], runner: Callable, **kwargs):
    assert isinstance(cmd, list), "run command must be a list"
    cmd_string = " ".join([shlex.quote(c) for c in cmd])
    # FIXME: This is gross to keep writing this, even when env is the same
    with tempfile.TemporaryFile(mode="w+") as run_file:
        print(assemble_env(), file=run_file)
        print("exec {}".format(cmd_string), file=run_file)
        run_file.flush()
        run_file.seek(0)
        try:
            return runner(["/bin/sh"], stdin=run_file, **kwargs)
        except subprocess.CalledProcessError as call_error:
            raise errors.SnapcraftCommandError(
                command=cmd_string, call_error=call_error
            ) from call_error


def run(cmd: List[str], **kwargs) -> None:
    _run(cmd, subprocess.check_call, **kwargs)


def run_output(cmd: List[str], **kwargs) -> str:
    output = _run(cmd, subprocess.check_output, **kwargs)
    try:
        return output.decode(sys.getfilesystemencoding()).strip()
    except UnicodeEncodeError:
        logger.warning("Could not decode output for {!r} correctly".format(cmd))
        return output.decode("latin-1", "surrogateescape").strip()


def get_core_path(base):
    """Returns the path to the core base snap."""
    return os.path.join(os.path.sep, "snap", base, "current")


def format_snap_name(snap, *, allow_empty_version: bool = False) -> str:
    """Return a filename representing the snap depending on snap attributes.

    :param dict snap: a dictionary of keys containing name, version and arch.
    :param bool allow_empty_version: if set a filename without a version is
                                     allowed.
    """
    if allow_empty_version and snap.get("version") is None:
        extension = "{name}_{arch}.snap"
    else:
        extension = "{name}_{version}_{arch}.snap"

    if "arch" not in snap:
        snap["arch"] = snap.get("architectures", None)
    if not snap["arch"]:
        snap["arch"] = "all"
    elif len(snap["arch"]) == 1:
        snap["arch"] = snap["arch"][0]
    else:
        snap["arch"] = "multi"

    return extension.format(**snap)


def is_snap() -> bool:
    snap_name = os.environ.get("SNAP_NAME", "")
    is_snap = snap_name == "snapcraft"
    logger.debug(
        "snapcraft is running as a snap {!r}, "
        "SNAP_NAME set to {!r}".format(is_snap, snap_name)
    )
    return is_snap


def is_docker_instance() -> bool:
    return os.path.exists(_DOCKERENV_FILE)


def set_plugindir(plugindir):
    global _plugindir
    _plugindir = plugindir


def get_plugindir():
    return _plugindir


def set_schemadir(schemadir):
    global _schemadir
    _schemadir = schemadir


def get_schemadir():
    return _schemadir


def set_extensionsdir(extensionsdir):
    global _extensionsdir
    _extensionsdir = extensionsdir


def get_extensionsdir():
    return _extensionsdir


def get_arch_triplet():
    raise errors.PluginOutdatedError("use 'project.arch_triplet'")


def get_arch():
    raise errors.PluginOutdatedError("use 'project.deb_arch'")


def get_parallel_build_count():
    raise errors.PluginOutdatedError("use 'parallel_build_count'")


def set_legacy_snapcraft_dir(snapcraftdir):
    global _legacy_snapcraft_dir
    _legacy_snapcraft_dir = snapcraftdir


def get_legacy_snapcraft_dir():
    return _legacy_snapcraft_dir


def get_python2_path(root):
    """Return a valid PYTHONPATH or raise an exception."""
    python_paths = glob.glob(
        os.path.join(root, "usr", "lib", "python2*", "dist-packages")
    )
    try:
        return python_paths[0]
    except IndexError:
        raise errors.SnapcraftEnvironmentError(
            "PYTHONPATH cannot be set for {!r}".format(root)
        )


def get_url_scheme(url):
    return urllib.parse.urlparse(url).scheme


def isurl(url):
    return get_url_scheme(url) != ""


def reset_env():
    global env
    env = []


def get_terminal_width(max_width=MAX_CHARACTERS_WRAP):
    if os.isatty(1):
        width = shutil.get_terminal_size().columns
    else:
        width = MAX_CHARACTERS_WRAP
    if max_width:
        width = min(max_width, width)
    return width


def format_output_in_columns(
    elements_list, max_width=MAX_CHARACTERS_WRAP, num_col_spaces=2
):
    """Return a formatted list of strings ready to be printed line by line

    elements_list is the list of elements ready to be printed on the output
    max_width is the number of caracters the output shouldn't exceed
    num_col_spaces is the number of spaces set between 2 columns"""

    # First, try to get the starting point in term of number of lines
    total_num_chars = sum([len(elem) for elem in elements_list])
    num_lines = math.ceil(
        (total_num_chars + (len(elements_list) - 1) * num_col_spaces) / max_width
    )
    sep = " " * num_col_spaces

    candidate_output = []
    while not candidate_output:
        # dispatch elements in resulting list until num_lines
        for i, element in enumerate(elements_list):
            # for new columns, get the maximum width of this column
            if i % num_lines == 0:
                col_width = 0
                for j in range(i, i + num_lines):
                    # ignore non existing elements at the end
                    with suppress(IndexError):
                        col_width = max(len(elements_list[j]), col_width)

            if i < num_lines:
                candidate_output.append([])
            candidate_output[i % num_lines].append(element.ljust(col_width))

        # check that any line (like the first one) is still smaller than
        # max_width
        if len(sep.join(candidate_output[0])) > max_width:
            # reset and try with one more line
            num_lines += 1
            candidate_output = []

    result_output = []
    for i, line in enumerate(candidate_output):
        result_output.append(sep.join(candidate_output[i]))

    return result_output


def get_include_paths(root, arch_triplet):
    paths = [
        os.path.join(root, "include"),
        os.path.join(root, "usr", "include"),
        os.path.join(root, "include", arch_triplet),
        os.path.join(root, "usr", "include", arch_triplet),
    ]

    return [p for p in paths if os.path.exists(p)]


def get_library_paths(root, arch_triplet, existing_only=True):
    """Returns common library paths for a snap.

    If existing_only is set the paths returned must exist for
    the root that was set.
    """
    paths = [
        os.path.join(root, "lib"),
        os.path.join(root, "usr", "lib"),
        os.path.join(root, "lib", arch_triplet),
        os.path.join(root, "usr", "lib", arch_triplet),
    ]

    if existing_only:
        paths = [p for p in paths if os.path.exists(p)]

    return paths


def get_pkg_config_paths(root, arch_triplet):
    paths = [
        os.path.join(root, "lib", "pkgconfig"),
        os.path.join(root, "lib", arch_triplet, "pkgconfig"),
        os.path.join(root, "usr", "lib", "pkgconfig"),
        os.path.join(root, "usr", "lib", arch_triplet, "pkgconfig"),
        os.path.join(root, "usr", "share", "pkgconfig"),
        os.path.join(root, "usr", "local", "lib", "pkgconfig"),
        os.path.join(root, "usr", "local", "lib", arch_triplet, "pkgconfig"),
        os.path.join(root, "usr", "local", "share", "pkgconfig"),
    ]

    return [p for p in paths if os.path.exists(p)]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Any, Callable, Dict, List, Set

from snapcraft import project
from .errors import GrammarSyntaxError
from . import typing
from ._statement import Statement
from ._on import OnStatement
from ._to import ToStatement
from ._try import TryStatement
from ._compound import CompoundStatement

_ON_TO_CLAUSE_PATTERN = re.compile(r"(\Aon\s+\S+)\s+(to\s+\S+\Z)")
_ON_CLAUSE_PATTERN = re.compile(r"\Aon\s+")
_TO_CLAUSE_PATTERN = re.compile(r"\Ato\s+")
_TRY_CLAUSE_PATTERN = re.compile(r"\Atry\Z")
_ELSE_CLAUSE_PATTERN = re.compile(r"\Aelse\Z")
_ELSE_FAIL_PATTERN = re.compile(r"\Aelse\s+fail\Z")


class GrammarProcessor:
    """The GrammarProcessor extracts desired primitives from grammar."""

    def __init__(
        self,
        grammar: typing.Grammar,
        project: project.Project,
        checker: Callable[[str], bool],
        *,
        transformer: Callable[[List[Statement], str, project.Project], str] = None
    ) -> None:
        """Create a new GrammarProcessor.

        :param list grammar: Unprocessed grammar.
        :param project: Instance of Project to use to determine appropriate
                        primitives.
        :type project: snapcraft.project.Project
        :param callable checker: callable accepting a single primitive,
                                 returning true if it is valid.
        :param callable transformer: callable accepting a call stack, single
                                     primitive, and project, and returning a
                                     transformed primitive.
        """
        self._grammar = grammar
        self.project = project
        self.checker = checker

        if transformer:
            self._transformer = transformer
        else:
            # By default, no transformation
            self._transformer = lambda s, p, o: p

    def process(
        self, *, grammar: typing.Grammar = None, call_stack: typing.CallStack = None
    ):
        """Process grammar and extract desired primitives.

        :param list grammar: Unprocessed grammar (defaults to that set in
                             init).
        :param list call_stack: Call stack of statements leading to now.

        :return: Primitives selected
        :rtype: set
        """

        if grammar is None:
            grammar = self._grammar

        if call_stack is None:
            call_stack = []

        primitives = set()  # type: Set[str]
        statements = _StatementCollection()
        statement = None  # type: Statement

        for section in grammar:
            if isinstance(section, str):
                # If the section is just a string, it's either "else fail" or a
                # primitive name.
                if _ELSE_FAIL_PATTERN.match(section):
                    _handle_else(statement, None)
                else:
                    primitives.add(self._transformer(call_stack, section, self.project))
            elif isinstance(section, dict):
                statement = self._parse_dict(section, statement, statements, call_stack)
            else:
                # jsonschema should never let us get here.
                raise GrammarSyntaxError(
                    "expected grammar section to be either of type 'str' or "
                    "type 'dict', but got {!r}".format(type(section))
                )

        # We've parsed the entire grammar, time to process it.
        statements.add(statement)
        primitives |= statements.process_all()

        return primitives

    def _parse_dict(
        self,
        section: Dict[str, Any],
        statement: Statement,
        statements: "_StatementCollection",
        call_stack: typing.CallStack,
    ):
        for key, value in section.items():
            # Grammar is always written as a list of selectors but the value
            # can be a list or a string. In the latter case we wrap it so no
            # special care needs to be taken when fetching the result from the
            # primitive.
            if not isinstance(value, list):
                value = {value}

            if _ON_TO_CLAUSE_PATTERN.match(key):
                # We've come across the beginning of a compound statement
                # with both 'on' and 'to'.

                # First, extract each statement's part of the string
                on, to = _ON_TO_CLAUSE_PATTERN.match(key).groups()

                # Now create a list of statements, in order
                compound_statements = [
                    OnStatement(
                        on=on, body=None, processor=self, call_stack=call_stack
                    ),
                    ToStatement(
                        to=to, body=None, processor=self, call_stack=call_stack
                    ),
                ]

                # Now our statement is a compound statement
                statement = CompoundStatement(
                    statements=compound_statements,
                    body=value,
                    processor=self,
                    call_stack=call_stack,
                )

            elif _ON_CLAUSE_PATTERN.match(key):
                # We've come across the beginning of an 'on' statement.
                # That means any previous statement we found is complete.
                # The first time through this may be None, but the
                # collection will ignore it.
                statements.add(statement)

                statement = OnStatement(
                    on=key, body=value, processor=self, call_stack=call_stack
                )

            elif _TO_CLAUSE_PATTERN.match(key):
                # We've come across the beginning of a 'to' statement.
                # That means any previous statement we found is complete.
                # The first time through this may be None, but the
                # collection will ignore it.
                statements.add(statement)

                statement = ToStatement(
                    to=key, body=value, processor=self, call_stack=call_stack
                )

            elif _TRY_CLAUSE_PATTERN.match(key):
                # We've come across the beginning of a 'try' statement.
                # That means any previous statement we found is complete.
                # The first time through this may be None, but the
                # collection will ignore it.
                statements.add(statement)

                statement = TryStatement(
                    body=value, processor=self, call_stack=call_stack
                )

            elif _ELSE_CLAUSE_PATTERN.match(key):
                _handle_else(statement, value)

        return statement


def _handle_else(statement: Statement, else_body: typing.Grammar):
    """Add else body to current statement.

    :param statement: The currently-active statement. If None it will be
                      ignored.
    :param else_body: The body of the else clause to add.

    :raises GrammarSyntaxError: If there isn't a currently-active
                                     statement.
    """

    try:
        statement.add_else(else_body)
    except AttributeError:
        raise GrammarSyntaxError(
            "'else' doesn't seem to correspond to an 'on' or 'try'"
        )


class _StatementCollection:
    """Unique collection of statements to run at a later time."""

    def __init__(self) -> None:
        self._statements = []  # type: List[Statement]

    def add(self, statement: Statement) -> None:
        """Add new statement to collection.

        :param statement: New statement.

        :raises GrammarSyntaxError: If statement is already in collection.
        """

        if not statement:
            return

        if statement in self._statements:
            raise GrammarSyntaxError(
                "found duplicate {!r} statements. These should be "
                "merged.".format(statement)
            )

        self._statements.append(statement)

    def process_all(self) -> Set[str]:
        """Process all statements in collection.

        :return: Selected primitives as judged by all statements in collection.
        :rtype: set
        """
        primitives = set()  # type: Set[str]
        for statement in self._statements:
            primitives |= statement.process()

        return primitives



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft import project
from snapcraft.internal.project_loader.grammar import (
    CompoundStatement,
    Statement,
    ToStatement,
    typing,
)


def _is_or_contains_to_statement(statement: Statement) -> bool:
    # If the statement is a ToStatement, then return True.
    if isinstance(statement, ToStatement):
        return True

    # If the statement is a CompoundStatement, check if a ToStatement is
    # contained within its statements. If so, return True.
    if isinstance(statement, CompoundStatement):
        return any(isinstance(s, ToStatement) for s in statement.statements)

    return False


def package_transformer(
    call_stack: typing.CallStack, package_name: str, project: project.Project
) -> str:
    if any(_is_or_contains_to_statement(s) for s in call_stack):
        if ":" not in package_name:
            # deb_arch is target arch or host arch if both are the same
            package_name += ":{}".format(project.deb_arch)

    return package_name



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from . import typing
from ._statement import Statement

# Don't use circular imports unless type checking
if TYPE_CHECKING:
    from ._processor import GrammarProcessor  # noqa: F401


class TryStatement(Statement):
    """Process a 'try' statement in the grammar.

    For example:
    >>> from snapcraft import ProjectOptions
    >>> from ._processor import GrammarProcessor
    >>> def checker(primitive):
    ...     return 'invalid' not in primitive
    >>> options = ProjectOptions()
    >>> processor = GrammarProcessor(None, options, checker)
    >>> clause = TryStatement(body=['invalid'], processor=processor)
    >>> clause.add_else(['valid'])
    >>> clause.process()
    {'valid'}
    """

    def __init__(
        self,
        *,
        body: typing.Grammar,
        processor: "GrammarProcessor",
        call_stack: typing.CallStack = None
    ) -> None:
        """Create a TryStatement instance.

        :param list body: The body of the clause.
        :param GrammarProcessor process: GrammarProcessor to use for processing
                                         this statement.
        :param list call_stack: Call stack leading to this statement.
        """
        super().__init__(
            body=body, processor=processor, call_stack=call_stack, check_primitives=True
        )

    def _check(self) -> bool:
        """Check if a statement main body should be processed.

        :return: True if main body should be processed, False if elses should
                 be processed.
        :rtype: bool
        """
        return self._validate_primitives(self._process_body())

    def __eq__(self, other) -> bool:
        return False

    def __str__(self) -> str:
        return "try"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities to run inside a correct snapcrafting environment."""


import tempfile

from snapcraft.internal import common


def which(command, **kwargs):
    """Returns the result of `which` run with the correct environment."""
    return run_script("which {}".format(command), **kwargs)


def getenv(envvar, **kwargs):
    """Return the specified envrionment variable."""
    return run_script("echo ${}".format(envvar), **kwargs)


def run_script(script, **kwargs):
    """Run the script and return the output."""
    with tempfile.NamedTemporaryFile("w+") as tempf:
        print("#!/bin/sh", file=tempf)
        print(script, file=tempf)
        tempf.flush()
        return common.run_output(["/bin/sh", tempf.name], **kwargs)



import contextlib
import os
from typing import Dict

from ._client import Client

from . import errors
from . import logger, _macaroon_auth
from . import constants


class SnapIndexClient(Client):
    """The Click Package Index knows everything about existing snaps.
    https://wiki.ubuntu.com/AppStore/Interfaces/ClickPackageIndex is the
    canonical reference.
    """

    def __init__(self, conf):
        """Initialize the SnapIndexClient object.

        :param config conf: Configuration details for the client.
        :type config: snapcraft.config.Config
        """
        super().__init__(
            conf,
            os.environ.get(
                "UBUNTU_STORE_SEARCH_ROOT_URL", constants.UBUNTU_STORE_SEARCH_ROOT_URL
            ),
        )

    def get_default_headers(self):
        """Return default headers for CPI requests.
        Tries to build an 'Authorization' header with local credentials
        if they are available.
        Also pin specific branded store if `SNAPCRAFT_UBUNTU_STORE`
        environment is set.
        """
        headers = {}

        with contextlib.suppress(errors.InvalidCredentialsError):
            headers["Authorization"] = _macaroon_auth(self.conf)

        branded_store = os.getenv("SNAPCRAFT_UBUNTU_STORE")
        if branded_store:
            headers["X-Ubuntu-Store"] = branded_store

        return headers

    def get_package(self, snap_name, channel=None, arch=None):
        """Get the details for the specified snap.

        :param str snap_name: Name of the snap.
        :param str channel: Channel of the snap.
        :param str arch: Architecture of the snap (none by default).

        :return Details for the snap.
        :rtype: dict
        """
        headers = self.get_default_headers()
        headers.update(
            {
                "Accept": "application/hal+json",
                "X-Ubuntu-Series": constants.DEFAULT_SERIES,
            }
        )
        if arch:
            headers["X-Ubuntu-Architecture"] = arch

        params = {
            # FIXME LP: #1662665
            "fields": "status,confinement,anon_download_url,download_url,"
            "download_sha3_384,download_sha512,snap_id,"
            "revision,release"
        }
        if channel is not None:
            params["channel"] = channel
        logger.debug("Getting details for {}".format(snap_name))
        url = "api/v1/snaps/details/{}".format(snap_name)
        resp = self.get(url, headers=headers, params=params)
        if resp.status_code != 200:
            raise errors.SnapNotFoundError(snap_name, channel, arch)
        return resp.json()

    def get_assertion(
        self, assertion_type: str, snap_id: str
    ) -> Dict[str, Dict[str, str]]:
        """Get the assertion for the specified snap.

        :param str assertion_type: The type of the assertion.
        :param str snap_id: The ID of the snap.

        :return Assertion for the snap.
        """
        headers = self.get_default_headers()
        logger.debug("Getting snap-declaration for {}".format(snap_id))
        url = "/api/v1/snaps/assertions/{}/{}/{}".format(
            assertion_type, constants.DEFAULT_SERIES, snap_id
        )
        response = self.get(url, headers=headers)
        if response.status_code != 200:
            raise errors.SnapNotFoundError(snap_id, series=constants.DEFAULT_SERIES)
        return response.json()

    def get(self, url, headers=None, params=None, stream=False):
        """Perform a GET request with the given arguments.

        :param str url: URL to send the request.
        :param dict headers: Headers to be sent along with the request.
        :param dict params: Query parameters to be sent along with
        the request.
        :param bool stream: Determines if the request shouldn't be
        automatically closed (true by default).

        :return Response of the request.
        """
        if headers is None:
            headers = self.get_default_headers()
        response = self.request(
            "GET", url, stream=stream, headers=headers, params=params
        )
        return response



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Facilities to output to the terminal.

These methods, which are named after common logging levels, wrap around
click.echo adding the corresponding color codes for each level.
"""
import click

from snapcraft.internal import common


def wrapped(msg: str) -> None:
    """Output msg wrapped to the terminal width to stdout.

    The maximum wrapping is determined by
    snapcraft.internal.common.MAX_CHARACTERS_WRAP
    """
    click.echo(
        click.formatting.wrap_text(
            msg, width=common.MAX_CHARACTERS_WRAP, preserve_paragraphs=True
        )
    )


def info(msg: str) -> None:
    """Output msg as informative to stdout.
    If the terminal supports colors the output will be green.
    """
    click.echo("\033[0;32m{}\033[0m".format(msg))


def warning(msg: str) -> None:
    """Output msg as a warning to stdout.
    If the terminal supports color the output will be yellow.
    """
    click.echo("\033[1;33m{}\033[0m".format(msg))


def error(msg: str) -> None:
    """Output msg as an error to stdout.
    If the terminal supports color the output will be red.
    """
    click.echo("\033[0;31m{}\033[0m".format(msg))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 James Beedy <jamesbeedy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The ruby plugin is useful for ruby based parts.
The plugin uses gem to install dependencies from a `Gemfile`.
This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.
Additionally, this plugin uses the following plugin-specific keywords:
    - gems:
      (list)
      A list of gems to install.
    - use-bundler
      (boolean)
      Use bundler to install gems from a Gemfile (defaults 'false').
    - ruby-version:
      (string)
      The version of ruby you want this snap to run.
"""

import glob
import logging
import os
import re

from snapcraft import BasePlugin, file_utils
from snapcraft.internal import errors
from snapcraft.sources import Tar


logger = logging.getLogger(__name__)


class RubyPlugin(BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["use-bundler"] = {"type": "boolean", "default": False}
        schema["properties"]["ruby-version"] = {"type": "string", "default": "2.4.2"}
        schema["properties"]["gems"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]
        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["ruby-version", "gems", "use-bundler"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        if project.info.base not in ("core16", "core18"):
            raise errors.PluginBaseError(part_name=self.name, base=project.info.base)

        # Beta Warning
        # Remove this comment and warning once ruby plugin is stable.
        logger.warn(
            "The ruby plugin is currently in beta, "
            "its API may break. Use at your own risk"
        )

        self._ruby_version = options.ruby_version
        self._ruby_part_dir = os.path.join(self.partdir, "ruby")
        self._ruby_download_url = "https://cache.ruby-lang.org/pub/ruby/ruby-{}.tar.gz".format(
            self._ruby_version
        )
        self._ruby_tar = Tar(self._ruby_download_url, self._ruby_part_dir)
        self._gems = options.gems or []
        self.build_packages.extend(
            ["gcc", "g++", "make", "zlib1g-dev", "libssl-dev", "libreadline-dev"]
        )

    def snap_fileset(self):
        fileset = super().snap_fileset()
        fileset.append("-include/")
        fileset.append("-share/")
        return fileset

    def pull(self):
        super().pull()
        os.makedirs(self._ruby_part_dir, exist_ok=True)

        logger.info("Fetching ruby {}...".format(self._ruby_version))
        self._ruby_tar.download()

        logger.info("Building/installing ruby...")
        self._ruby_install(builddir=self._ruby_part_dir)

        self._gem_install()
        if self.options.use_bundler:
            self._bundle_install()

    def env(self, root):
        env = super().env(root)

        for key, value in self._env_dict(root).items():
            env.append('{}="{}"'.format(key, value))

        return env

    def _env_dict(self, root):
        env = dict()
        rubydir = os.path.join(root, "lib", "ruby")

        # Patch versions of ruby continue to use the minor version's RUBYLIB,
        # GEM_HOME, and GEM_PATH. Fortunately there should just be one, so we
        # can detect it by globbing instead of trying to determine what the
        # minor version is programmatically
        versions = glob.glob(os.path.join(rubydir, "gems", "*"))

        # Before Ruby has been pulled/installed, no versions will be found.
        # If that's the case, we won't define any Ruby-specific variables yet
        if len(versions) == 1:
            ruby_version = os.path.basename(versions[0])

            rubylib = os.path.join(rubydir, ruby_version)

            # Ruby uses some pretty convoluted rules for determining its
            # arch-specific RUBYLIB. Rather than try and duplicate that logic
            # here, let's just look for a file that we know is in there:
            # rbconfig.rb. There should only be one.
            paths = glob.glob(os.path.join(rubylib, "*", "rbconfig.rb"))
            if len(paths) != 1:
                raise errors.SnapcraftEnvironmentError(
                    "Expected a single rbconfig.rb, but found {}".format(len(paths))
                )

            env["RUBYLIB"] = "{}:{}".format(rubylib, os.path.dirname(paths[0]))
            env["GEM_HOME"] = os.path.join(rubydir, "gems", ruby_version)
            env["GEM_PATH"] = os.path.join(rubydir, "gems", ruby_version)
        elif len(versions) > 1:
            raise errors.SnapcraftEnvironmentError(
                "Expected a single Ruby version, but found {}".format(len(versions))
            )

        return env

    def _run(self, command, **kwargs):
        """Regenerate the build environment, then run requested command.

        Without this function, the build environment would not be regenerated
        and thus the newly installed Ruby would not be discovered.
        """

        env = os.environ.copy()
        env.update(self._env_dict(self.installdir))
        self.run(command, env=env, **kwargs)

    def _ruby_install(self, builddir):
        self._ruby_tar.provision(builddir, clean_target=False, keep_tarball=True)
        self._run(["./configure", "--disable-install-rdoc", "--prefix=/"], cwd=builddir)
        self._run(["make", "-j{}".format(self.parallel_build_count)], cwd=builddir)
        self._run(
            ["make", "install", "DESTDIR={}".format(self.installdir)], cwd=builddir
        )
        # Fix all shebangs to use the in-snap ruby
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"^#!.*ruby"),
            r"#!/usr/bin/env ruby",
        )

    def _gem_install(self):
        if self.options.use_bundler:
            self._gems = self._gems + ["bundler"]
        if self._gems:
            logger.info("Installing gems...")
            gem_install_cmd = [
                os.path.join(self.installdir, "bin", "ruby"),
                os.path.join(self.installdir, "bin", "gem"),
                "install",
                "--env-shebang",
            ]
            self._run(gem_install_cmd + self._gems)

    def _bundle_install(self):
        bundle_install_cmd = [
            os.path.join(self.installdir, "bin", "ruby"),
            os.path.join(self.installdir, "bin", "bundle"),
            "install",
        ]
        self._run(bundle_install_cmd)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import subprocess
from copy import deepcopy
from datetime import datetime

from . import StoreClient
from . import errors
from . import constants


class _BaseAssertion:
    """Private Base class to handle assertions.

    Implementations are supposed to define a class attribute to determine
    the assertion type.

    :cvar str _assertion_type: the assertion type, also treated as the endpoint
                               for the assertion on the store and the payload
                               header key for the returned data.
    """

    @property
    def publisher_id(self):
        """Return the publisher-id of a snap.
        This entry is also known as accound-id or developer-id.

        The value is lazily fetched from the store.
        """
        if not self._account_info:
            self._account_info = self._store_client.get_account_information()
        return self._account_info["account_id"]

    @property
    def snap_id(self):
        """Return the snap-id of the given snap_name.

        The value is lazily fetched from the store.
        """
        if not self._account_info:
            self._account_info = self._store_client.get_account_information()
        snaps = self._account_info["snaps"][self._release]
        try:
            return snaps[self._snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(self._snap_name)

    def __init__(self, *, snap_name, signing_key=None):
        """Create an instance to handle an assertion.

        :param str snap_name: snap name to handle assertion for.
        :param str signing_key: the name of the key to use, if not
                                provided, the default key is used.
        :ivar dict assertion: holds the actual assertion.
        :ivar dict signed_assertion: holds a signed version of assertion.
        """
        self._store_client = StoreClient()
        self._snap_name = snap_name
        self._signing_key = signing_key
        self._release = constants.DEFAULT_SERIES
        self._account_info = None
        self.assertion = None
        self.signed_assertion = None

    def get(self):
        """Get the current assertion from the store.

        :returns: the assertion corresponding to the snap_name.
        :rtype: dict
        """
        # The store adds a header key which is not consistent with the endpoint
        # which we need to pop as it is not understood by snap sign.
        if self.assertion:
            return self.assertion
        store_assertion = self._store_client.get_assertion(
            self.snap_id, self._assertion_type
        )
        self.assertion = list(store_assertion.values())[0]
        return self.assertion

    def sign(self):
        """Create a signed version of the obtained assertion.

        :returns: signed assertion document.
        """
        cmdline = ["snap", "sign"]
        if self._signing_key:
            cmdline += ["-k", self._signing_key]
        snap_sign = subprocess.Popen(
            cmdline,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        data = json.dumps(self.assertion).encode("utf8")
        assertion, err = snap_sign.communicate(input=data)
        if snap_sign.returncode != 0:
            err = err.decode("ascii", errors="replace")
            raise errors.StoreAssertionError(
                snap_name=self._snap_name, endpoint=self._assertion_type, error=err
            )

        self.signed_assertion = assertion
        return assertion

    def push(self, force=False):
        """Push the assertion to the store, signing if necessary.

        :param bool force: if True, ignore any conflict with revoked developers
                           and the snap revisions it would invalidate.
        :returns: None
        """
        if not self.signed_assertion:
            self.sign()
        self._store_client.push_assertion(
            self.snap_id, self.signed_assertion, self._assertion_type, force=force
        )


class DeveloperAssertion(_BaseAssertion):
    """Implementation of a developer assertion.
    The assertion is used to enable collaboration for a given snap_name
    by updating a snap-id's developer assertion using the store endpoint.

    The assertion content has the following structure
    {
        'type': 'snap-developer',
        'authority-id': '<account_id of the publisher or store authority-id>',
        'publisher-id': '<account_id of the publisher>',
        'snap-id': '<snap-id>',
        'developers': [{
            'developer-id': '<account-id>',
            'since': '2017-02-10T08:35:00.390258Z'
           },{
            'developer-id': '<account-id>',
            'since': '2017-02-10T08:35:00.390258Z',
            'until': '2018-02-10T08:35:00.390258Z'
           }],
        }

    """

    _assertion_type = "developers"

    def new_assertion(self, *, developers):
        """Create a new assertion with developers based out of the current one.

        The new assertion has its assertion's authority-id normalized to the
        assertion's publisher_id and the assertion's revision increased by 1.

        :param list developers: a list of a dictionary developers holding the
                                keys: developer_id (mandatory), since
                                (mandatory) and until (optional).
        :returns: a new assertion based out of the current assertion with the
                  provided developers list.
        :rtype: DeveloperAssertion
        """
        new_assertion = deepcopy(self.assertion)
        new_assertion["developers"] = developers

        # The revision should be incremented, to avoid `invalid-revision`
        # errors.
        new_assertion["revision"] = str(int(self.assertion.get("revision", "-1")))

        # There is a possibility that the `authority-id` to be `canonical`,
        # which should be changed to the `publisher_id` to match the signing
        # key.
        new_assertion["authority-id"] = self.publisher_id

        new_instance = DeveloperAssertion(
            snap_name=self._snap_name, signing_key=self._signing_key
        )
        # Reference the already fetched information
        new_instance._account_info = self._account_info
        new_instance.assertion = new_assertion
        return new_instance

    def get(self):
        """Return a dict containing the developer assertion for snap_name.

        The data that comes from the store query looks as follows:
        {'snap_developer': {
             'type': 'snap-developer',
             'authority-id': <account_id of the publisher or
                              store authority-id>,
             'publisher-id': <account_id of the publisher>,
             'snap-id': 'snap_id',
             'developers': [{
                 'developer-id': 'account_id of dev-1',
                 'since': '2017-02-10T08:35:00.390258Z'
                },{
                 'developer-id': 'account_id of dev-2',
                 'since': '2017-02-10T08:35:00.390258Z',
                 'until': '2018-02-10T08:35:00.390258Z'
                }],
             }
        }

        The assertion is saved without the snap_developer header.

        :returns: the latest developer assertion corresponding to snap_name.
        :rtype: dict
        """
        try:
            self.assertion = super().get()
        except errors.StoreValidationError as e:
            if e.error_list[0]["code"] != "snap-developer-not-found":
                raise
            self.assertion = {
                "type": "snap-developer",
                "authority-id": self.publisher_id,
                "publisher-id": self.publisher_id,
                "snap-id": self.snap_id,
                "developers": [],
            }
        # A safeguard to operate easily on the assertion
        if "developers" not in self.assertion:
            self.assertion["developers"] = []
        return self.assertion

    def get_developers(self):
        """Return a copy of the current developers listed in the assertion.

        :returns: a list of developers.
        :rtype: list
        """
        return self.get()["developers"].copy()

    def is_equal(self, other_assertion):
        """Determine equality of developer lists in each assertion.

        During the comparison, differences in milliseconds are not considered.

        :returns: Return True if the list of developers in this instances
                  assertion list is equal to the list from other_assertion.
        :rtype: bool
        """
        this_devs = self._normalize_time(self.assertion["developers"].copy())
        other_devs = self._normalize_time(
            other_assertion.assertion["developers"].copy()
        )
        return this_devs == other_devs

    def _normalize_time(self, developers):
        for dev in developers:
            for range_ in ["since", "until"]:
                if range_ in dev:
                    date = datetime.strptime(dev[range_], "%Y-%m-%dT%H:%M:%S.%fZ")
                    dev[range_] = date.strftime("%Y-%m-%dT%H:%M:%S.000000Z")
        return developers



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import logging
import sys
from subprocess import check_call, check_output, CalledProcessError
from typing import Sequence
from urllib import parse

import requests_unixsocket
from requests import exceptions

from . import errors


_CHANNEL_RISKS = ["stable", "candidate", "beta", "edge"]
logger = logging.getLogger(__name__)


# TODO https://bugs.launchpad.net/snapcraft/+bug/1786868


class SnapPackage:
    """SnapPackage acts as a mediator to install or refresh a snap.

    It uses information provided by snapd implicitly referring to the local
    and remote stores to obtain information about the snap, such as its
    confinement value and channel availability.

    This information can also be used to determine if a snap should be
    installed or refreshed.

    There are risks of the data falling out of date between the query and the
    requested action given that it is not possible to hold a global lock on
    snapd and the store data can change in between validation and execution.
    """

    @classmethod
    def is_valid_snap(cls, snap):
        return cls(snap).is_valid()

    @classmethod
    def is_snap_installed(cls, snap):
        return cls(snap).installed

    def __init__(self, snap):
        """Lifecycle handler for a snap of the format <snap-name>/<channel>."""
        self.name, self.channel = _get_parsed_snap(snap)
        self._original_channel = self.channel
        if not self.channel or self.channel == "stable":
            self.channel = "latest/stable"

        # This store information from a local request
        self._local_snap_info = None
        # And this stores information from a remote request
        self._store_snap_info = None

        self._is_installed = None
        self._is_in_store = None

    @property
    def installed(self):
        if self._is_installed is None:
            self._is_installed = self.get_local_snap_info() is not None
        return self._is_installed

    @property
    def in_store(self):
        if self._is_in_store is None:
            try:
                self._is_in_store = self.get_store_snap_info() is not None
            except errors.SnapUnavailableError:
                self._is_in_store = False
        return self._is_in_store

    def get_local_snap_info(self):
        """Returns a local payload for the snap.

        Validity of the results are determined by checking self.installed."""
        if self._is_installed is None:
            with contextlib.suppress(exceptions.HTTPError):
                self._local_snap_info = _get_local_snap_info(self.name)
        return self._local_snap_info

    def get_store_snap_info(self):
        """Returns a store payload for the snap."""
        if self._is_in_store is None:
            # Some environments timeout often, like the armv7 testing
            # infrastructure. Given that constraint, we add some retry
            # logic.
            retry_count = 5
            while retry_count > 0:
                try:
                    self._store_snap_info = _get_store_snap_info(self.name)
                    break
                except exceptions.HTTPError as http_error:
                    logger.debug(
                        "The http error when checking the store for "
                        "{!r} is {!r} (retries left {})".format(
                            self.name, http_error.response.status_code, retry_count
                        )
                    )
                    if http_error.response.status_code == 404:
                        raise errors.SnapUnavailableError(
                            snap_name=self.name, snap_channel=self.channel
                        )
                    retry_count -= 1

        return self._store_snap_info

    def _get_store_channels(self):
        snap_store_info = self.get_store_snap_info()
        if not self.in_store:
            return dict()

        return snap_store_info["channels"]

    def get_current_channel(self):
        current_channel = ""
        if self.installed:
            local_snap_info = self.get_local_snap_info()
            current_channel = local_snap_info["channel"]
            if any([current_channel.startswith(risk) for risk in _CHANNEL_RISKS]):
                current_channel = "latest/{}".format(current_channel)
        return current_channel

    def is_classic(self):
        store_channels = self._get_store_channels()
        try:
            return store_channels[self.channel]["confinement"] == "classic"
        except KeyError:
            # We have seen some KeyError issues when running tests that are
            # hard to debug as they only occur there, logging in debug mode
            # will help uncover the root cause if it happens again.
            logger.debug(
                "Current store channels are {!r} and the store"
                "payload is {!r}".format(store_channels, self._store_snap_info)
            )
            raise

    def is_valid(self):
        """Check if the snap is valid."""
        if not self.in_store:
            return False
        store_channels = self._get_store_channels()
        return self.channel in store_channels.keys()

    def install(self):
        """Installs the snap onto the system."""
        snap_install_cmd = []
        if _snap_command_requires_sudo():
            snap_install_cmd = ["sudo"]
        snap_install_cmd.extend(["snap", "install", self.name])
        if self._original_channel:
            snap_install_cmd.extend(["--channel", self._original_channel])
        if self.is_classic():
            # TODO make this a user explicit choice
            snap_install_cmd.append("--classic")
        try:
            check_call(snap_install_cmd)
        except CalledProcessError as install_error:
            raise errors.SnapInstallError(
                snap_name=self.name, snap_channel=self.channel
            )

    def refresh(self):
        """Refreshes a snap onto a channel on the system."""
        snap_refresh_cmd = []
        if _snap_command_requires_sudo():
            snap_refresh_cmd = ["sudo"]
        snap_refresh_cmd.extend(
            ["snap", "refresh", self.name, "--channel", self.channel]
        )
        if self.is_classic():
            # TODO make this a user explicit choice
            snap_refresh_cmd.append("--classic")
        try:
            check_call(snap_refresh_cmd)
        except CalledProcessError as install_error:
            raise errors.SnapRefreshError(
                snap_name=self.name, snap_channel=self.channel
            )


def install_snaps(snaps_list):
    """Install snaps of the format <snap-name>/<channel>.

    :return: a list of "name=revision" for the snaps installed.
    """
    snaps_installed = []
    for snap in snaps_list:
        snap_pkg = SnapPackage(snap)
        if not snap_pkg.is_valid():
            raise errors.SnapUnavailableError(
                snap_name=snap_pkg.name, snap_channel=snap_pkg.channel
            )

        if not snap_pkg.installed:
            snap_pkg.install()
        elif snap_pkg.get_current_channel() != snap_pkg.channel:
            snap_pkg.refresh()

        snap_pkg = SnapPackage(snap)
        snaps_installed.append(
            "{}={}".format(snap_pkg.name, snap_pkg.get_local_snap_info()["revision"])
        )
    return snaps_installed


def _snap_command_requires_sudo():
    # snap whoami returns - if the user is not logged in.
    output = check_output(["snap", "whoami"])
    whoami = output.decode(sys.getfilesystemencoding())
    requires_root = False
    try:
        requires_root = whoami.split(":")[1].strip() == "-"
    # A safeguard if the output changes
    except IndexError:
        requires_root = True
    if requires_root:
        logger.warning("snapd is not logged in, snap install commands will use sudo")
    return requires_root


def get_assertion(assertion_params: Sequence[str]) -> bytes:
    """Get assertion information.

    :param assertion_params: a sequence of strings to pass to 'snap known'.
    :returns: a stream of bytes from the assertion.
    :rtype: bytes
    """
    try:
        return check_output(["snap", "known", *assertion_params])
    except CalledProcessError as call_error:
        raise errors.SnapGetAssertionError(
            assertion_params=assertion_params
        ) from call_error


def _get_parsed_snap(snap):
    if "/" in snap:
        sep_index = snap.find("/")
        snap_name = snap[:sep_index]
        snap_channel = snap[sep_index + 1 :]
    else:
        snap_name = snap
        snap_channel = ""
    return snap_name, snap_channel


def get_snapd_socket_path_template():
    return "http+unix://%2Frun%2Fsnapd.socket/v2/{}"


def _get_local_snap_info(snap_name):
    slug = "snaps/{}".format(parse.quote(snap_name, safe=""))
    url = get_snapd_socket_path_template().format(slug)
    try:
        snap_info = requests_unixsocket.get(url)
    except exceptions.ConnectionError as e:
        raise errors.SnapdConnectionError(snap_name, url) from e
    snap_info.raise_for_status()
    return snap_info.json()["result"]


def _get_store_snap_info(snap_name):
    # This logic uses /v2/find returns an array of results, given that
    # we do a strict search either 1 result or a 404 will be returned.
    slug = "find?{}".format(parse.urlencode(dict(name=snap_name)))
    url = get_snapd_socket_path_template().format(slug)
    snap_info = requests_unixsocket.get(url)
    snap_info.raise_for_status()
    return snap_info.json()["result"][0]


def get_installed_snaps():
    """Return all the snaps installed in the system.

    :return: a list of "name=revision" for the snaps installed.
    """
    slug = "snaps"
    url = get_snapd_socket_path_template().format(slug)
    try:
        snap_info = requests_unixsocket.get(url)
        snap_info.raise_for_status()
        local_snaps = snap_info.json()["result"]
    except exceptions.ConnectionError as e:
        local_snaps = []
    return ["{}={}".format(snap["name"], snap["revision"]) for snap in local_snaps]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import logging
import os
import sys
import re
import shutil
from typing import Any, Dict

from snapcraft.project import Project
from snapcraft.internal import project_loader
from . import errors

logger = logging.getLogger(__name__)


_EXPECTED_SNAP_DIR_PATTERNS = {
    re.compile(r"^snapcraft.yaml$"),
    re.compile(r"^.snapcraft(/state)?$"),
    re.compile(r"^hooks(/.*)?$"),
    re.compile(r"^local(/.*)?$"),
    re.compile(r"^plugins(/.*)?$"),
    re.compile(r"^gui(/.*\.(png|svg|desktop))?$"),
}


def conduct_build_environment_sanity_check(provider: str):
    if provider == "multipass":
        _check_multipass_installed()


def _check_multipass_installed():
    if shutil.which("multipass"):
        return

    if sys.platform != "linux":
        raise errors.MultipassMissingNonLinuxError()

    if not shutil.which("snap"):
        raise errors.SnapMissingLinuxError()
    else:
        raise errors.MultipassMissingLinuxError()


def conduct_project_sanity_check(project: Project) -> None:
    """Sanity check the project itself before continuing.

    The checks done here are meant to be light, and not rely on the build environment.
    """
    # The snapcraft.yaml should be valid even without extensions applied
    # This here check is mostly for backwards compatibility with the
    # rest of the code base.
    if project.info is not None:
        project.info.validate_raw_snapcraft()

    snap_dir_path = os.path.join(project._project_dir, "snap")
    if os.path.isdir(snap_dir_path):
        _check_snap_dir(snap_dir_path)


def conduct_environment_sanity_check(
    project: Project, yaml_data: Dict[str, Any], schema: Dict[str, Any]
) -> None:
    """Sanity check the build environment and expanded YAML.

    :param snapcraft.Project project: Project settings
    :param dict yaml_data: Validated YAML, with extensions applied
    """
    # Never modify the YAML data
    yaml_data = copy.deepcopy(yaml_data)

    _verify_command_chain_only_full_adapter(yaml_data, schema)


def _check_snap_dir(snap_dir_path: str) -> None:
    unexpected_paths = set()
    for root, directories, files in os.walk(snap_dir_path):
        for entry in directories + files:
            path = os.path.relpath(os.path.join(root, entry), snap_dir_path)
            if not _snap_dir_path_expected(path):
                unexpected_paths.add(path)

    if unexpected_paths:
        logger.warn(
            "The snap/ directory is meant specifically for snapcraft, but it contains "
            "the following non-snapcraft-related paths, which is unsupported and will "
            "cause unexpected behavior:"
            "\n- {}\n\n"
            "If you must store these files within the snap/ directory, move them to "
            "snap/local/, which is ignored by snapcraft.".format(
                "\n- ".join(sorted(unexpected_paths))
            )
        )


def _snap_dir_path_expected(path: str) -> bool:
    for pattern in _EXPECTED_SNAP_DIR_PATTERNS:
        if pattern.match(path):
            return True
    return False


def _verify_command_chain_only_full_adapter(
    yaml_data: Dict[str, Any], schema: Dict[str, Any]
) -> None:
    # Determine the default adapter
    app_schema = schema["apps"]["patternProperties"]["^[a-zA-Z0-9](?:-?[a-zA-Z0-9])*$"][
        "properties"
    ]
    default_adapter = app_schema["adapter"]["default"]

    # Loop through all apps
    for app_name, app_definition in yaml_data.get("apps", dict()).items():
        # Verify that, if command-chain is used, the "full" adapter is also used
        if "command-chain" in app_definition:
            adapter_string = app_definition.get("adapter", default_adapter)
            adapter = project_loader.Adapter[adapter_string.upper()]
            if adapter == project_loader.Adapter.LEGACY:
                raise errors.CommandChainWithLegacyAdapterError(app_name)



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import os
import re
import sys


def recursive_data_files(directory, install_directory):
    data_files = []
    for root, directories, file_names in os.walk(directory):
        file_paths = [os.path.join(root, file_name) for file_name in file_names]
        data_files.append((os.path.join(install_directory, root), file_paths))
    return data_files


# Common distribution data
name = "snapcraft"
version = "devel"
description = "Easily craft snaps from multiple sources"
author_email = "snapcraft@lists.snapcraft.io"
url = "https://github.com/snapcore/snapcraft"
packages = [
    "snapcraft",
    "snapcraft.cli",
    "snapcraft.cli.snapcraftctl",
    "snapcraft.extractors",
    "snapcraft.integrations",
    "snapcraft.internal",
    "snapcraft.internal.cache",
    "snapcraft.internal.build_providers",
    "snapcraft.internal.build_providers._multipass",
    "snapcraft.internal.deltas",
    "snapcraft.internal.lifecycle",
    "snapcraft.internal.meta",
    "snapcraft.internal.pluginhandler",
    "snapcraft.internal.project_loader",
    "snapcraft.internal.project_loader.grammar",
    "snapcraft.internal.project_loader.grammar_processing",
    "snapcraft.internal.project_loader.inspection",
    "snapcraft.internal.project_loader._extensions",
    "snapcraft.internal.repo",
    "snapcraft.internal.sources",
    "snapcraft.internal.states",
    "snapcraft.project",
    "snapcraft.plugins",
    "snapcraft.plugins._ros",
    "snapcraft.plugins._python",
    "snapcraft.storeapi",
]
package_data = {"snapcraft.internal.repo": ["manifest.txt"]}
license = "GPL v3"
classifiers = (
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Natural Language :: English",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Topic :: Software Development :: Build Tools",
    "Topic :: System :: Software Distribution",
)

# look/set what version we have
changelog = "debian/changelog"
if os.path.exists(changelog):
    head = codecs.open(changelog, encoding="utf-8").readline()
    match = re.compile(".*\((.*)\).*").match(head)
    if match:
        version = match.group(1)


# If on Windows, construct an exe distribution
if sys.platform == "win32":
    from cx_Freeze import setup, Executable

    # cx_Freeze relevant options
    build_exe_options = {
        # Explicitly add any missed packages that are not found at runtime
        "packages": [
            "pkg_resources",
            "pymacaroons",
            "click",
            "responses",
            "configparser",
            "cffi",
        ],
        # Explicit inclusion data, which is then clobbered.
        "include_files": [
            ("schema", os.path.join("share", "snapcraft", "schema")),
            ("extensions", os.path.join("share", "snapcraft", "extensions")),
        ],
    }

    exe = Executable(script="bin/snapcraft", base=None)  # console subsystem

    setup(
        name=name,
        version=version,
        description=description,
        author_email=author_email,
        url=url,
        packages=packages,
        package_data=package_data,
        license=license,
        classifiers=classifiers,
        # cx_Freeze-specific arguments
        options={"build_exe": build_exe_options},
        executables=[exe],
    )

# On other platforms, continue as normal
else:
    from setuptools import setup

    setup(
        name=name,
        version=version,
        description=description,
        author_email=author_email,
        url=url,
        packages=packages,
        package_data=package_data,
        license=license,
        classifiers=classifiers,
        entry_points=dict(console_scripts=["snapcraft = snapcraft.cli.__main__:run"]),
        # snapcraftctl is not in console_scripts because we need a clean environment.
        scripts=["bin/snapcraftctl"],
        data_files=[
            ("share/snapcraft/schema", ["schema/" + x for x in os.listdir("schema")])
        ]
        + recursive_data_files("extensions", "share/snapcraft"),
        install_requires=["pysha3", "pyxdg", "requests"],
        test_suite="tests.unit",
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from snapcraft.internal import lifecycle, steps
from snapcraft.internal.project_loader import _config


def lifecycle_status(config: _config.Config) -> List[Dict[str, str]]:
    """Return a list of dicts summarizing the current lifecycle status.

    :param _config.Config config: Config object representing loaded project status.
    :returns: List of dicts summarizing the current lifecycle status (perfect for
              printing via tabulate).
    """
    cache = lifecycle.StatusCache(config)

    summary = []
    for part in config.all_parts:
        part_summary = {"part": part.name}
        for step in steps.STEPS:
            part_summary[step.name] = None
            if cache.has_step_run(part, step):
                part_summary[step.name] = "complete"

            dirty_report = cache.get_dirty_report(part, step)
            if dirty_report:
                part_summary[step.name] = "dirty ({})".format(
                    dirty_report.get_summary()
                )

            outdated_report = cache.get_outdated_report(part, step)
            if outdated_report:
                part_summary[step.name] = "outdated ({})".format(
                    outdated_report.get_summary()
                )
        summary.append(part_summary)

    return summary



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat

from ._base import FileBase


class Script(FileBase):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
    ):
        super().__init__(
            source, source_dir, source_tag, source_commit, source_branch, source_depth
        )

    def download(self):
        super().download()
        st = os.stat(self.file)
        os.chmod(self.file, st.st_mode | stat.S_IEXEC)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Tim Süberkrüb
# Copyright (C) 2018 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import tempfile

from . import errors
from ._base import FileBase


class SevenZip(FileBase):
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
            "7zip",
        )
        if source_tag:
            raise errors.SnapcraftSourceInvalidOptionError("7z", "source-tag")
        elif source_commit:
            raise errors.SnapcraftSourceInvalidOptionError("7z", "source-commit")
        elif source_branch:
            raise errors.SnapcraftSourceInvalidOptionError("7z", "source-branch")

    def provision(self, dst, clean_target=True, keep_7z=False, src=None):
        if src:
            seven_zip_file = src
        else:
            seven_zip_file = os.path.join(
                self.source_dir, os.path.basename(self.source)
            )
        seven_zip_file = os.path.realpath(seven_zip_file)

        if clean_target:
            tmp_7z = tempfile.NamedTemporaryFile().name
            shutil.move(seven_zip_file, tmp_7z)
            shutil.rmtree(dst)
            os.makedirs(dst)
            shutil.move(tmp_7z, seven_zip_file)

        self._run_output(["7z", "x", seven_zip_file], cwd=dst)

        if not keep_7z:
            os.remove(seven_zip_file)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Create parts containing a Plainbox test collection known as a provider.

Plainbox is a toolkit consisting of python3 library, development tools,
documentation and examples. It is targeted at developers working on testing or
certification applications and authors creating tests for such applications.
More information: http://plainbox.readthedocs.org/en/latest/

To find out more about authoring a plainbox provider, see the following
documentation: http://plainbox.readthedocs.org/en/latest/author/providers.html

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.
"""

import os

import snapcraft
from snapcraft.internal import errors
from snapcraft.internal import mangling


class PlainboxProviderPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["required"] = ["source"]
        return schema

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_base_tools(project.info.base)

    def _setup_base_tools(self, base):
        if base in ("core16", "core18"):
            self.stage_packages.extend(
                ["python3-pip", "python3-wheel", "python3-setuptools"]
            )
            self.build_packages.append("intltool")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def build(self):
        super().build()

        env = os.environ.copy()
        # Ensure the first provider does not attempt to validate against
        # providers installed on the build host by initialising PROVIDERPATH
        # to empty
        env["PROVIDERPATH"] = ""
        provider_stage_dir = os.path.join(self.project.stage_dir, "providers")
        if os.path.exists(provider_stage_dir):
            provider_dirs = [
                os.path.join(provider_stage_dir, provider)
                for provider in os.listdir(provider_stage_dir)
            ]
            env["PROVIDERPATH"] = ":".join(provider_dirs)
        self.run(["python3", "manage.py", "validate"], env=env)
        self.run(["python3", "manage.py", "build"])
        self.run(["python3", "manage.py", "i18n"])
        self.run(
            [
                "python3",
                "manage.py",
                "install",
                "--layout=relocatable",
                "--prefix=/providers/{}".format(self.name),
                "--root={}".format(self.installdir),
            ]
        )

        mangling.rewrite_python_shebangs(self.installdir)

    def snap_fileset(self):
        fileset = super().snap_fileset()
        # If a python package is added as a stage-packages it will include
        # sitecustomize.py which is irrelevant and will cause unnecessary
        # conflicts so instead we just ignore these entries.
        fileset.append("-usr/lib/python*/sitecustomize.py")
        fileset.append("-etc/python*/sitecustomize.py")
        return fileset



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from subprocess import CalledProcessError
from typing import Dict, List, Union

from snapcraft import formatting_utils
from snapcraft.internal import steps


class SnapcraftError(Exception):
    """Base class for all snapcraft exceptions.

    :cvar fmt: A format string that daughter classes override

    """

    fmt = "Daughter classes should redefine this"

    def __init__(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return self.fmt.format([], **self.__dict__)

    def get_exit_code(self):
        """Exit code to use if this exception causes Snapcraft to exit."""
        return 2


class SnapcraftReportableError(SnapcraftError):
    """Base class for all snapcraft exceptions that integrate with Sentry.

    An exception class inheriting from this will get a prompt to report
    to sentry if raised and is handled by the exception handler on program
    exit.

    :cvar fmt: A format string that daughter classes override

    """


class MissingStateCleanError(SnapcraftError):
    fmt = (
        "Failed to clean: "
        "Missing state for {step.name!r}. "
        "To clean the project, run `snapcraft clean`."
    )

    def __init__(self, step):
        super().__init__(step=step)


class StepOutdatedError(SnapcraftError):

    fmt = (
        "Failed to reuse files from previous run: "
        "The {step.name!r} step of {part!r} is out of date:\n"
        "{report}"
        "To continue, clean that part's {step.name!r} step by running "
        "`snapcraft clean {parts_names} -s {step.name}`."
    )

    def __init__(
        self, *, step, part, dirty_report=None, outdated_report=None, dependents=None
    ):
        messages = []

        if dirty_report:
            messages.append(dirty_report.get_report())

        if outdated_report:
            messages.append(outdated_report.get_report())

        if dependents:
            humanized_dependents = formatting_utils.humanize_list(dependents, "and")
            pluralized_dependents = formatting_utils.pluralize(
                dependents, "depends", "depend"
            )
            messages.append(
                "The {0!r} step for {1!r} needs to be run again, "
                "but {2} {3} on it.\n".format(
                    step.name, part, humanized_dependents, pluralized_dependents
                )
            )
            parts_names = ["{!s}".format(d) for d in sorted(dependents)]
        else:
            parts_names = [part]

        super().__init__(
            step=step,
            part=part,
            report="".join(messages),
            parts_names=" ".join(parts_names),
        )


class SnapcraftEnvironmentError(SnapcraftError):
    # FIXME This exception is too generic.
    # https://bugs.launchpad.net/snapcraft/+bug/1734231
    # --elopio - 20171123

    fmt = "{message}"

    def __init__(self, message):
        super().__init__(message=message)


class SnapcraftMissingLinkerInBaseError(SnapcraftError):

    fmt = (
        "Cannot find the linker to use for the target base {base!r}.\n"
        "Please verify that the linker exists at the expected path "
        "{linker_path!r} and try again. If the linker does not exist "
        "contact the author of the base (run `snap info {base}` to get "
        "information for this base)."
    )

    def __init__(self, *, base, linker_path):
        super().__init__(base=base, linker_path=linker_path)


class IncompatibleBaseError(SnapcraftError):
    fmt = (
        "The linker version {linker_version!r} used by the base {base!r} is "
        "incompatible with files in this snap:\n"
        "{file_list}"
    )

    def __init__(
        self, *, base: str, linker_version: str, file_list: Dict[str, str]
    ) -> None:
        spaced_file_list = ("    {} ({})".format(k, v) for k, v in file_list.items())
        super().__init__(
            base=base,
            linker_version=linker_version,
            file_list="\n".join(sorted(spaced_file_list)),
        )


class PrimeFileConflictError(SnapcraftError):

    fmt = (
        "Failed to filter files: "
        "The following files have been excluded by the `stage` keyword, "
        "but included by the `prime` keyword: {fileset!r}. "
        "Edit the `snapcraft.yaml` to make sure that the files included in "
        "`prime` are also included in `stage`."
    )


class InvalidAppCommandError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "The specified command {command!r} defined in the app {app_name!r} does "
        "not exist or is not executable.\n"
        "Ensure that {command!r} is relative to the prime directory."
    )

    def __init__(self, command, app_name):
        super().__init__(command=command, app_name=app_name)


class InvalidAppCommandFormatError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "The specified command {command!r} defined in the app {app_name!r} does "
        "not match the pattern expected by snapd.\n"
        "The command must consist only of alphanumeric characters, spaces, and the "
        "following special characters: / . _ # : $ -"
    )

    def __init__(self, command, app_name):
        super().__init__(command=command, app_name=app_name)


class InvalidCommandChainError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "The command-chain item {item!r} defined in the app {app_name!r} does "
        "not exist or is not executable.\n"
        "Ensure that {item!r} is relative to the prime directory."
    )

    def __init__(self, item, app_name):
        super().__init__(item=item, app_name=app_name)


class InvalidDesktopFileError(SnapcraftError):

    fmt = (
        "Failed to generate desktop file: "
        "Invalid desktop file {filename!r}: {message}."
        # FIXME include how to fix each of the possible desktop file errors.
        # https://bugs.launchpad.net/snapcraft/+bug/1727435
        # --elopio - 2017-10-25
    )

    def __init__(self, filename, message):
        super().__init__(filename=filename, message=message)


class PluginError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "{message}"
        # FIXME include how to fix each of the possible plugin errors.
        # https://bugs.launchpad.net/snapcraft/+bug/1727484
        # --elopio - 2017-10-25
    )

    def __init__(self, message):
        super().__init__(message=message)


class PluginBaseError(SnapcraftError):

    fmt = "The plugin used by part {part_name!r} does not support snaps using base {base!r}."

    def __init__(self, *, part_name, base):
        super().__init__(part_name=part_name, base=base)


class SnapcraftPartConflictError(SnapcraftError):

    fmt = (
        "Failed to stage: "
        "Parts {other_part_name!r} and {part_name!r} have the following "
        "files, but with different contents:\n"
        "{file_paths}\n\n"
        "Snapcraft offers some capabilities to solve this by use of the "
        "following keywords:\n"
        "    - `filesets`\n"
        "    - `stage`\n"
        "    - `snap`\n"
        "    - `organize`\n\n"
        "To learn more about these part keywords, run "
        "`snapcraft help plugins`."
    )

    def __init__(self, *, part_name, other_part_name, conflict_files):
        spaced_conflict_files = ("    {}".format(i) for i in conflict_files)
        super().__init__(
            part_name=part_name,
            other_part_name=other_part_name,
            file_paths="\n".join(sorted(spaced_conflict_files)),
        )


class SnapcraftOrganizeError(SnapcraftError):

    fmt = "Failed to organize part {part_name!r}: {message}"

    def __init__(self, part_name, message):
        super().__init__(part_name=part_name, message=message)


class MissingCommandError(SnapcraftError):

    fmt = (
        "Failed to run command: "
        "One or more packages are missing, please install:"
        " {required_commands!r}"
    )

    def __init__(self, required_commands):
        super().__init__(required_commands=required_commands)


class InvalidWikiEntryError(SnapcraftError):

    fmt = (
        "Invalid wiki entry: "
        "{error!r}"
        # FIXME include how to fix each of the possible wiki errors.
        # https://bugs.launchpad.net/snapcraft/+bug/1727490
        # --elopio - 2017-10-25
    )

    def __init__(self, error=None):
        super().__init__(error=error)


class MissingGadgetError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "Missing gadget.yaml file.\n"
        "When creating gadget snaps you are required to provide a gadget.yaml "
        "file\n"
        "in the root of your snapcraft project\n\n"
        "Read more about gadget snaps and the gadget.yaml on:\n"
        "https://github.com/snapcore/snapd/wiki/Gadget-snap"
    )


class PluginOutdatedError(SnapcraftError):

    fmt = "This plugin is outdated: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class ToolMissingError(SnapcraftReportableError):

    fmt = (
        "A tool snapcraft depends on could not be found: {command_name!r}.\n"
        "Ensure the tool is installed and available, and try again."
    )

    def __init__(self, *, command_name: str) -> None:
        super().__init__(command_name=command_name)


class RequiredCommandFailure(SnapcraftError):

    fmt = "{command!r} failed."


class RequiredCommandNotFound(SnapcraftError):

    fmt = "{cmd_list[0]!r} not found."


class RequiredPathDoesNotExist(SnapcraftError):

    fmt = "Required path does not exist: {path!r}"


class SnapcraftPathEntryError(SnapcraftError):

    fmt = (
        "Failed to generate snap metadata: "
        "The path {value!r} set for {key!r} in {app!r} does not exist. "
        "Make sure that the files are in the `prime` directory."
    )


class InvalidPullPropertiesError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "Invalid pull properties specified by {plugin_name!r} plugin: "
        "{properties}"
        # TODO add a link to some docs about pull properties?
        # --elopio - 2017-10-25
    )

    def __init__(self, plugin_name, properties):
        super().__init__(plugin_name=plugin_name, properties=properties)


class InvalidBuildPropertiesError(SnapcraftError):

    fmt = (
        "Failed to load plugin: "
        "Invalid build properties specified by {plugin_name!r} plugin: "
        "{properties}"
        # TODO add a link to some docs about build properties?
        # --elopio - 2017-10-25
    )

    def __init__(self, plugin_name, properties):
        super().__init__(plugin_name=plugin_name, properties=properties)


class StagePackageDownloadError(SnapcraftError):

    fmt = (
        "Failed to fetch stage packages: "
        "Error downloading packages for part {part_name!r}: {message}."
    )

    def __init__(self, part_name, message):
        super().__init__(part_name=part_name, message=message)


class OsReleaseIdError(SnapcraftError):

    fmt = "Unable to determine host OS ID"


class OsReleaseNameError(SnapcraftError):

    fmt = "Unable to determine host OS name"


class OsReleaseVersionIdError(SnapcraftError):

    fmt = "Unable to determine host OS version ID"


class OsReleaseCodenameError(SnapcraftError):

    fmt = "Unable to determine host OS version codename"


class InvalidContainerImageInfoError(SnapcraftError):

    fmt = (
        "Failed to parse container image info: "
        "SNAPCRAFT_IMAGE_INFO is not a valid JSON string: {image_info}"
    )

    def __init__(self, image_info: str) -> None:
        super().__init__(image_info=image_info)


class PatcherError(SnapcraftError):
    pass


class PatcherGenericError(PatcherError):

    fmt = (
        "{elf_file!r} cannot be patched to function properly in a classic "
        "confined snap: {message}"
    )

    def __init__(self, *, elf_file, process_exception):
        message = "{} failed with exit code {}".format(
            " ".join(process_exception.cmd), process_exception.returncode
        )
        super().__init__(elf_file=elf_file, message=message)


class PatcherNewerPatchelfError(PatcherError):

    fmt = (
        "{elf_file!r} cannot be patched to function properly in a classic "
        "confined snap: {message}.\n"
        "{patchelf_version!r} may be too old. A newer version of patchelf "
        "may be required.\n"
        "Try adding the `after: [patchelf]` and a `patchelf` part that would "
        "filter out files from prime `prime: [-*]` or "
        "`build-snaps: [patchelf/latest/edge]` to the failing part in your "
        "`snapcraft.yaml` to use a newer patchelf."
    )

    def __init__(self, *, elf_file, process_exception, patchelf_version):
        message = "{} failed with exit code {}".format(
            " ".join(process_exception.cmd), process_exception.returncode
        )
        super().__init__(
            elf_file=elf_file, message=message, patchelf_version=patchelf_version
        )


class StagePackageMissingError(SnapcraftError):

    fmt = (
        "{package!r} is required inside the snap for this part to work "
        "properly.\n"
        "Add it as a `stage-packages` entry for this part."
    )

    def __init__(self, *, package):
        super().__init__(package=package)


class MetadataExtractionError(SnapcraftError):
    pass


class MissingMetadataFileError(MetadataExtractionError):

    fmt = (
        "Failed to generate snap metadata: "
        "Part {part_name!r} has a 'parse-info' referring to metadata file "
        "{path!r}, which does not exist."
    )

    def __init__(self, part_name: str, path: str) -> None:
        super().__init__(part_name=part_name, path=path)


class UnhandledMetadataFileTypeError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "This type of file is not supported for supplying metadata."
    )

    def __init__(self, path: str) -> None:
        super().__init__(path=path)


class InvalidExtractorValueError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "Extractor {extractor_name!r} didn't return ExtractedMetadata as "
        "expected."
    )

    def __init__(self, path: str, extractor_name: str) -> None:
        super().__init__(path=path, extractor_name=extractor_name)


class SnapcraftCommandError(SnapcraftError, CalledProcessError):
    """Exception for generic command errors.

    Processes should capture this error for specific messaging.
    This exception carries the signature of CalledProcessError for backwards
    compatibility.
    """

    fmt = "Failed to run {command!r}: Exited with code {exit_code}."

    def __init__(self, *, command: str, call_error: CalledProcessError) -> None:
        super().__init__(command=command, exit_code=call_error.returncode)
        CalledProcessError.__init__(
            self,
            returncode=call_error.returncode,
            cmd=call_error.cmd,
            output=call_error.output,
            stderr=call_error.stderr,
        )


class SnapcraftPluginCommandError(SnapcraftError):
    """Command executed by a plugin fails."""

    fmt = (
        "Failed to run {command!r} for {part_name!r}: "
        "Exited with code {exit_code}.\n"
        "Verify that the part is using the correct parameters and try again."
    )

    def __init__(
        self, *, command: Union[List, str], part_name: str, exit_code: int
    ) -> None:
        if isinstance(command, list):
            command = " ".join(command)
        super().__init__(command=command, part_name=part_name, exit_code=exit_code)


class ScriptletBaseError(SnapcraftError):
    """Base class for all scriptlet-related exceptions.

    :cvar fmt: A format string that daughter classes override

    """


class ScriptletRunError(ScriptletBaseError):
    fmt = "Failed to run {scriptlet_name!r}: Exit code was {code}."

    def __init__(self, scriptlet_name: str, code: int) -> None:
        super().__init__(scriptlet_name=scriptlet_name, code=code)


class ScriptletDuplicateDataError(ScriptletBaseError):
    fmt = (
        "Failed to save data from scriptlet into {step.name!r} state: "
        "The {humanized_keys} key(s) were already saved in the "
        "{other_step.name!r} step."
    )

    def __init__(
        self, step: steps.Step, other_step: steps.Step, keys: List[str]
    ) -> None:
        self.keys = keys
        super().__init__(
            step=step,
            other_step=other_step,
            humanized_keys=formatting_utils.humanize_list(keys, "and"),
        )


class ScriptletDuplicateFieldError(ScriptletBaseError):
    fmt = "Unable to set {field}: it was already set in the {step.name!r} step."

    def __init__(self, field: str, step: steps.Step) -> None:
        super().__init__(field=field, step=step)


class SnapcraftctlError(ScriptletBaseError):
    fmt = "{message}"

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class SnapcraftInvalidCLIConfigError(SnapcraftError):

    fmt = "The cli configuration file {config_file!r} has invalid data: {error!r}."

    def __init__(self, *, config_file: str, error: str) -> None:
        super().__init__(config_file=config_file, error=error)
        # This is to keep mypy happy
        self.config_file = config_file


class SnapcraftCopyFileNotFoundError(SnapcraftError):
    fmt = (
        "Failed to copy {path!r}: no such file or directory.\n"
        "Check the path and try again."
    )

    def __init__(self, path):
        super().__init__(path=path)


class InvalidStepError(SnapcraftError):
    fmt = "{step_name!r} is not a valid lifecycle step"

    def __init__(self, step_name):
        super().__init__(step_name=step_name)


class StepHasNotRunError(SnapcraftError):
    fmt = "The {part_name!r} part has not yet run the {step.name!r} step"

    def __init__(self, part_name, step):
        super().__init__(part_name=part_name, step=step)


class NoLatestStepError(SnapcraftError):
    fmt = "The {part_name!r} part hasn't run any steps"

    def __init__(self, part_name):
        super().__init__(part_name=part_name)


class NoNextStepError(SnapcraftError):
    fmt = "The {part_name!r} part has run through its entire lifecycle"

    def __init__(self, part_name):
        super().__init__(part_name=part_name)


class MountPointNotFoundError(SnapcraftError):
    fmt = "Nothing is mounted at {mount_point!r}"

    def __init__(self, mount_point):
        super().__init__(mount_point=mount_point)


class RootNotMountedError(SnapcraftError):
    fmt = "{root!r} is not mounted"

    def __init__(self, root):
        super().__init__(root=root)


class InvalidMountinfoFormat(SnapcraftError):
    fmt = "Unable to parse mountinfo row: {row}"

    def __init__(self, row):
        super().__init__(row=row)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click

import snapcraft

SNAPCRAFT_VERSION_TEMPLATE = "snapcraft, version %(version)s"


@click.group()
def versioncli():
    """Version commands"""
    pass


@versioncli.command("version")
def version():
    """Obtain snapcraft's version number.

    Examples:
        snapcraft version
        snapcraft --version
    """
    click.echo(SNAPCRAFT_VERSION_TEMPLATE % {"version": snapcraft.__version__})



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import locale
import os
import subprocess
import sys

from ._runner import run as run_snapcraft
from .snapcraftctl._runner import run as run_snapcraftctl  # noqa
from .echo import warning
from snapcraft import project, yaml_utils
from snapcraft.internal import common

# If the locale ends up being ascii, Click will barf. Let's try to prevent that
# here by using C.UTF-8 as a last-resort fallback. This mostly happens in CI,
# using LXD or Docker. This is the same logic used by Click to error out.
if codecs.lookup(locale.getpreferredencoding()).name == "ascii" and os.name == "posix":
    output = subprocess.check_output(["locale", "-a"]).decode("ascii", "replace")

    for line in output.splitlines():
        this_locale = line.strip()
        if this_locale.lower() in ("c.utf8", "c.utf-8"):
            warning("Locale not set! Snapcraft will temporarily use C.UTF-8")
            os.environ["LC_ALL"] = "C.UTF-8"
            os.environ["LANG"] = "C.UTF-8"
            break


def _needs_legacy() -> bool:
    if not os.path.isdir(common.get_legacy_snapcraft_dir()):
        return False

    try:
        # Early bootstrapping does not allow us to use the existing utilities we
        # have to manage this check.
        if os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT") == "managed-host":
            base_dir = os.path.expanduser(os.path.join("~", "project"))
        else:
            base_dir = None
        snapcraft_yaml_path = project.get_snapcraft_yaml(base_dir=base_dir)
        with open(snapcraft_yaml_path, "r") as f:
            data = yaml_utils.load(f)
        return data.get("base") is None
    except Exception:
        # If there are issues loading/parsing the YAML, just pass off to the current version
        # where the error should be properly handled
        return False


def run_legacy_snapcraft() -> None:
    legacy_python = os.path.join(
        common.get_legacy_snapcraft_dir(), "usr", "bin", "python3"
    )
    legacy_snapcraft = os.path.join(
        common.get_legacy_snapcraft_dir(), "bin", "snapcraft"
    )

    os.execv(legacy_python, [legacy_python, legacy_snapcraft] + sys.argv[1:])


def run():
    if _needs_legacy():
        run_legacy_snapcraft()
    else:
        run_snapcraft(prog_name="snapcraft")


if __name__ == "__main__":
    run()
